-module(ksudoku).
-author(hanak@inf.bme.hu).
-vsn('2009-11-14').

% -compile(export_all).
-export([sudoku_be/1,sudoku_ki/2,megold/2,stopper/2]).

-import(sudoku,[sudoku/1]).

%% @type sspec() = {dist(), board()}.
%% @type dist()  = integer().
%% @type field() = [info()].
%% @type info()  = s | w | integer().
%% @type board() = [[field()]].

%% @type ssol() = [[integer()]].

%% @spec sudoku:sudoku(SSpec::sspec()) -> SSols::[ssol()].
%% @doc  SSols az SSpec feladványt kielégítő megoldások listája.

%% ksudoku-függvények

%% @spec ksudoku:sudoku_be(FileIn::string()) -> SSpec::sspec().
%% @doc  SSpec a FileIn szövegfájlból beolvasott Sudoku-feladvány.
%%
sudoku_be(FileIn) ->
    case file:open(FileIn,read) of
	{ok, InputFile} ->
	    {N,_} = case io:get_line(InputFile,'') of
			eof ->
			    io:format("Ures a(z) ~s fajl.~n",[FileIn]),
			    exit('Hibas feladvany');
			I ->
			    string:to_integer(cleaned(I))
		    end,
	    T = tabla_be(InputFile),
	    file:close(InputFile),
%	    io:fwrite("~w\n",[T]),
	    KK = length(T),
	    if N < 1 orelse KK =< N ->
		    io:format("A ~w tavolsagkonstans nem 1 es (sorhossz-1) kozotti ertek.~n",[N]),
		    exit('Hibas feladvany');
	       true ->
		    case helyes_feladvany(KK,T) of
			true ->
			    {N,T};
			false ->
			    io:format("Hibas a feladvany a(z) ~s fajlban.~n",[FileIn]),
			    exit('Hibas feladvany')
		    end
	    end;
	_ ->
	    io:format("Nincs ~s nevu fajl.~n",[FileIn]),
	    exit('Nincs ilyen feladvany')
    end.

tabla_be(InputFile) ->
    tabla_be(InputFile,[]).

tabla_be(InputFile,Zss) ->
    case io:get_line(InputFile,'') of
	eof ->
	    KK = length(Zss),
	    if KK < 1 orelse 25 < KK ->
		    io:format("1-nel kisebb vagy 5-nel nagyobb a ~w cellameret.~n",[math:sqrt(KK)]),
		    exit('Hibas feladvany');
	       true ->
		    case lists:all(fun(L) -> length(L) == KK end, Zss) of
			false ->
			    io:format("Nem negyzetes a feladvany, vagy nem egyforma hosszuak a sorok.~n",[]),
			    exit('Hibas feladvany');
			true ->
			    lists:reverse(Zss)
		    end
	    end;
	Sor ->
	    Ss = string:tokens(cleaned(Sor)," "),
	    Zs = lists:map(fun str2infolist/1, Ss),
	    case Zs of
		[] ->
		    tabla_be(InputFile,Zss);
		_ ->
		    tabla_be(InputFile,[Zs|Zss])
	    end
    end.

cleaned(Sor) ->
    S0 = lists:filter(fun(C) -> C /= $\t end,Sor),
    string:strip(string:strip(S0),right,$\n).

str2infolist([$-]) ->
    [];
str2infolist(XXs=[X|_Xs]) when $0=<X andalso X=<$9 ->
    {I,Rs} = string:to_integer(XXs),
    [I|str2infolist(Rs)];
str2infolist([C|Xs]) ->
    [list_to_atom([C])|str2infolist(Xs)];
str2infolist([]) ->
    [].

%% @spec helyes_feladvany(Size::integer(),Board::board()) -> B::bool().
%% @doc  B igaz, ha a feladvány megfelel a specifikációnak.
%%
helyes_feladvany(KK,T) ->
    lists:all(fun(Fs) -> helyes_mezo(KK,Fs) end, lists:concat(T)).

helyes_mezo(KK,Fs) ->
    helyes_mezo(KK,Fs,0,0,0).

helyes_mezo(_KK,[],S,V,W) ->
    S =< 1 andalso V =< 1 andalso W =< 1;
helyes_mezo(KK,[s|Fs],S,V,W) ->
    helyes_mezo(KK,Fs,S+1,V,W);
helyes_mezo(KK,[w|Fs],S,V,W) ->
    helyes_mezo(KK,Fs,S,V,W+1);
helyes_mezo(KK,[I|Fs],S,V,W) when is_integer(I) ->
    1 =< I andalso I =< KK andalso helyes_mezo(KK,Fs,S,V+1,W);
helyes_mezo(_KK,[I|_Fs],_S,_V,_W) ->
    io:format("Hibas ~w info van az input fajlban.~n",[I]),
    exit('Hibas feladvany').

%% @spec ksudoku:sudoku_ki(FileOut::string(),SSols::[ssol()]) -> void()
%% @doc  Az SSols listában átadott Sudoku-megoldásokat kiírja
%%       a FileOut szövegfájlba.
%%
sudoku_ki(FileOut,SSols) ->
    {ok,OutputFile} = file:open(FileOut,write),
    tablak_ki(OutputFile,SSols),
    file:close(OutputFile).

tablak_ki(OutputFile,[SSol]) ->
    io:fwrite(OutputFile,"~s~n",[tabla_ki(SSol)]);
tablak_ki(OutputFile,[SSol|SSols]) ->
    io:fwrite(OutputFile,"~s~n~n",[tabla_ki(SSol)]),
    tablak_ki(OutputFile,SSols);
tablak_ki(OutputFile,[]) ->
    io:fwrite(OutputFile,"",[]).

tabla_ki([L]) ->
    sor_ki(L);
tabla_ki([L|Ls]) ->
    sor_ki(L) ++ "\n" ++ tabla_ki(Ls);
tabla_ki(_) ->
    [].

sor_ki([H|T]) ->
    I2L = fun integer_to_list/1,
    I2L(H) ++ lists:foldr(fun(A,B) -> " " ++ I2L(A) ++ B end,"",T);
sor_ki([]) ->
    "".

%% @spec ksudoku:megold(FileIn::string(),FileOut::string()) -> void()
%% @doc  Beolvas egy feladványt a FileIn szövegfájlból és összes
%%       megoldását kiírja a FileOut szövegfájlba. Ehhez
%%       felhasználja a sudoku:sudoku/1 függvényt.
%%
megold(FileIn,FileOut) ->
    KT = sudoku_be(FileIn),
    SSols = sudoku:sudoku(KT), 
    sudoku_ki(FileOut,SSols).

%% @spec ksudoku:stopper(FileIn::string(),FileOut::string()) -> void()
%% @doc  Mint ksudoku:megold/2, de a végén kiírja a FileIn nevet,
%%       a megoldások számát és a futási idõt is.
%%
stopper(FileIn,FileOut) ->
    KT = sudoku_be(FileIn),
    {T, SSols} = timer:tc(sudoku,sudoku,[KT]),
    sudoku_ki(FileOut,SSols),
    io:format("Feladvany: ~s, megoldas: ~w, futasi ido: ~w ms\n",
	      [FileIn,length(SSols),T div 1000]).
